@@ -1,11 +1,11 @@
 /* Delete entries from a tar archive.
 
    Copyright (C) 1988, 1992, 1994, 1996, 1997, 2000, 2001, 2003, 2004,
-   2005, 2006 Free Software Foundation, Inc.
+   2005, 2006, 2010 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
-   Free Software Foundation; either version 2, or (at your option) any later
+   Free Software Foundation; either version 3, or (at your option) any later
    version.
 
    This program is distributed in the hope that it will be useful, but
@@ -35,7 +35,6 @@ extern union block *current_block;
 extern union block *recent_long_name;
 extern union block *recent_long_link;
 extern off_t records_read;
-extern off_t records_written;
 
 /* The number of records skipped at the start of the archive, when
    passing over members that are not deleted.  */
@@ -166,7 +165,9 @@ delete_archive_members (void)
 
   do
     {
-      enum read_header status = read_header (true);
+      enum read_header status = read_header (&current_header,
+                                             &current_stat_info,
+                                             read_header_x_raw);
 
       switch (status)
 	{
@@ -261,7 +262,8 @@ delete_archive_members (void)
 
 	  if (current_block == record_end)
 	    flush_archive ();
-	  status = read_header (false);
+	  status = read_header (&current_header, &current_stat_info, 
+	                        read_header_auto);
 
 	  xheader_decode (&current_stat_info);
 
@@ -307,10 +309,10 @@ delete_archive_members (void)
 	    }
 	  /* Copy header.  */
 
-	  if (extended_header.size)
+	  if (current_stat_info.xhdr.size)
 	    {
-	      write_recent_bytes (extended_header.buffer,
-				  extended_header.size);
+	      write_recent_bytes (current_stat_info.xhdr.buffer,
+				  current_stat_info.xhdr.size);
 	    }
 	  else
 	    {
